import 'dart:ffi';
import 'package:ffi/ffi.dart';

import 'array.dart';

/// Instead of doing many little allocations, [Arena]
/// allocates a large chunk of memory which can be sliced into multiple [Array]s.
/// When all operations have finished, the [Arena] can be deleted and it frees all
/// memory in one operation. This can significantly improve performance and it makes
/// memory management simpler when dealing with many allocations.
/// ````
/// final arena = Arena<Uint8>(10, Uint8Array.fromPtr);
/// final array = arena.newArray(5);
/// final array2 = arena.newArray(5);
/// void some_native_c_api(array.ptr, array.view.length, array2.ptr, array2.view.length);
/// arena.delete();
/// ````
class Arena<T extends NativeType> {
  int _index = 0;

  /// Total length of allocated memory
  final int length;
  final Pointer<T> _arrPtr;
  final Array<T> Function(Pointer<T>, int) _arrayFromPtr;

  /// Left free space of Arena.
  int get freeSpace => length - _index;

  /// [_arrayFromPtr] constructs a new [Array] with [length] elements
  Arena(this.length, this._arrayFromPtr) : _arrPtr = allocate(count: length);

  /// [_arrPtr] points to a pre-allocated array of [length].
  Arena.fromPtr(this._arrPtr, this.length, this._arrayFromPtr);

  /// Returns a slice of [length] elements from the pre-allocated array of Arena.
  /// When [length] exceeds [freeSpace] it throws a [RangeError].
  Array<T> newArray(int length) {
    assert(length > 0);
    if (freeSpace < length) {
      throw RangeError('Not enough free space');
    }
    final ptrAtIndex = _arrPtr.elementAt(_index);
    final arr = _arrayFromPtr(ptrAtIndex, length);
    _index += length;
    return arr;
  }

  /// Deletes all allocated memory.
  void delete() {
    free(_arrPtr);
  }
}

extension Uint8Arena on Arena<Uint8> {
  static Arena<Uint8> allocate(int length) => Arena(length, Uint8Array.fromPtr);
  static Arena<Uint8> fromPtr(Pointer<Uint8> ptr, int length) =>
      Arena.fromPtr(ptr, length, Uint8Array.fromPtr);
}
