import 'dart:ffi';
import 'package:ffi/ffi.dart';
import 'package:meta/meta.dart';
import 'dart:typed_data';
import 'array_utils.dart';

/// Thin wrapper around a [Pointer] and its associated typed list.
/// For a more type safe and convenient method to construct [Array] have
/// a look at the extensions like [Uint8Array].
@sealed
class Array<T extends NativeType> {
  @doNotStore
  final Pointer<T> ptr;

  List<int> _view;

  @doNotStore
  List<int> get untypedView => _view;

  /// [_view] is the associated typed list of [ptr]
  Array.fromPtr(this.ptr, this._view);

  /// [getView] returns the associated typed list of the [Pointer] pointing to the array.
  Array(int length, List<int> Function(Pointer<T> ptr, int length) getView)
      : ptr = allocate<T>(count: length) {
    _view = getView(ptr, length);
  }

  Array.fromList(
      List<int> list, List<int> Function(Pointer<T> ptr, int length) getView)
      : ptr = allocate<T>(count: list.length) {
    _view = getView(ptr, list.length)..setAll(0, list);
  }

  void delete() {
    free(ptr);
  }

  @override
  bool operator ==(dynamic arr) {
    if (arr is Pointer) {
      return arr == ptr;
    }
    if (arr is Array) {
      return arr.ptr == ptr;
    }
    return false;
  }

  @override
  int get hashCode => ptr.hashCode;

  @override
  String toString() {
    return '${runtimeType}($ptr, $_view)';
  }
}

extension Uint8Array on Array<Uint8> {
  Uint8List get view => untypedView;
  static Array<Uint8> fromList(Uint8List list) =>
      Array.fromList(list, getUint8View);

  static Array<Uint8> allocate(int length) => Array(length, getUint8View);
  static Array<Uint8> fromPtr(Pointer<Uint8> ptr, int length) =>
      Array.fromPtr(ptr, ptr.asTypedList(length));
}
